/* Emperor - an X windows strategy game for 1-6 players
   Copyright (C) 1998 Frank Rustemeyer.
   This file is part of the Emperor package.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

   If you have changes and you want them to be included in a future
   release of Emperor, please e-mail them to mambres@mud.tap.de.

   This file declares the classes for the game logic and should
   be (and remain) independent of the visualisation.

   98/11/30, Frank Rustemeyer <mambres@mud.tap.de>: Released.
   99/01/29, Frank Rustemeyer <mambres@mud.tap.de>: Added roads.
   99/01/29, Frank Rustemeyer <mambres@mud.tap.de>: Added language support.
   99/02/09, Frank Rustemeyer <mambres@mud.tap.de>: Fixed and straightened the computer move
   99/02/11, Frank Rustemeyer <mambres@mud.tap.de>: Major rewrite of the computer player
*/

#ifndef EMPGAME
#define EMPGAME

#include <stdio.h>
#include "emp.h"
#include "text.h"

class coordinate
{   public:
       int x;
       int y;
       void set(int sx, int sy) { x = sx; y = sy; };
       coordinate operator + (coordinate &c);
       int operator - (coordinate &c); // distance between two points
       int operator == (const coordinate &c) const;
       int operator != (const coordinate &c) const;
    };

class list
  // a linked list. members will be added by giving the list
  // as an argument to the listmember constructor
  // the list also has functions for being displayed in the main window 
{   public:
	 friend class listmember;
	list(void);
	int init(class listmember * member);
	  // initialises list when member is the first member to be added
	  // returns zero if there are already members in the list
	void remove(class listmember * member);
	  // remove and destruct a member
	int length(void) const { return len; };
	  // returns the length of the list
	int position(void) const { return pos; };
	  // returns the position of the root pointer
	void save(FILE *f);
	  // save list.
	void load(FILE *f, int flag);
	  // load list.
	class listmember * root;
	  // pointer to listmember at position pos.
        class listmember *find(coordinate c);
    protected:
	int before(class listmember * member) const;
	  // checks if member appears in the list before root
	int len;  // current length of the list
	int pos;  // current position of root
    };

class listmember  // generic member of a list (abstract base class)
{   public:
	listmember(list & thislist, listmember *aft = NULL);
	  // create a member in thislist (after member aft, or at the end)
	listmember(void);                  // create member outside list
        virtual ~listmember(void) {}
	  // delete a member. should be called ONLY from list::remove
	  // to make sure list::length is adjusted
        void unlink(void);
        void addtolist(list & thislist, listmember *aft = NULL);
        // insert this listmember into thislist (after member aft, or at the end)
	listmember * previous(void) const { return prv; };
	  // the previous list member
	listmember * next(void) const { return nxt; };
	  // the next list member
	listmember * first(void) const;    // the first member in the list
	listmember * last(void) const;     // the last member in the list
        listmember * find(coordinate c) const;
	virtual void save(FILE *f) = 0;
	  // save this listmember to a file, called by list::save
	virtual void load(char *s) = 0;
	  // load this listmember from a string read from a file,
          // called by list::load
	coordinate getpos(void) { return pos; };
        void setpos(coordinate c) { pos.set(c.x, c.y); };
	char getowner()         { return owner; };
	void setowner(char o);
    protected:
	coordinate pos;
        char owner;
    private:
	listmember * prv;  // pointer to previous listmember
	listmember * nxt;  // pointer to next listmember
    };

class matrix
{   public:
	matrix();
	int & operator [] (coordinate c) { return f[c.x + SX*c.y]; };
        void save(FILE *f);
        void load(FILE *f);
    protected:
	int f[SX*SY];
    };

class map : public matrix
{   public:
	map();
    private:
	void growth(coordinate c);
    };

class plan : public matrix
{   public:
        plan();
        void calculate(int add);
        void calcair(void);
        void settargets(int mode);
        void addrandom(void);
        void markownisles(void);
        void calcwater(void);
        void clearmap(void);
        void blankblockedfields(void);
        coordinate select_target_field(coordinate unit_position,
                                       int water_only = 0);
    private:
        coordinate linksoben, rechtsunten;
    };

class city : public listmember
{   public:
	city(list & thislist);
        city();
	char prodtime() 	{ return ptime; };
	void setprodtime(char p){ ptime = p; };
	int produce(char o);
	void setprod(int typ);
	virtual void save(FILE *f);
	virtual void load(char *s);
        class unit * produced;
    private:
	char ptime;
	};

class unit : public listmember
{   public:
	unit();
        unit(list & thislist);
        virtual ~unit() {}

	char * prodtimestring(char * s);
        char * hitpointstring(char * s);
        virtual char * restmovestring(char * s);
        int restmoves(void) { return canmove() - hasmoved; }
	void setpos(coordinate c);
        coordinate movetowards(coordinate m);
        int clearmoves(int o);
        void countmove(void) { hasmoved++; }
        void nomoves(void) {hasmoved = canmove();}
        void hit(void) { damage++; }
        void kill(void);
        char * repair(char *s);
        class unit *alive(void);
        class unit * findnext(int player , int start);
        virtual int validsquare(coordinate c) = 0;
        int checkcombat(coordinate c);
        virtual int id(void) = 0;
	virtual char * name(void) = 0;
	virtual char prodtime(void) = 0;
        virtual int repairtime(void) = 0;
        virtual int maxdamage(void) = 0;
        virtual int movement_cost(coordinate dest) { return 1; }
        virtual int canmove(void) = 0;
	virtual void save(FILE *f);
	virtual void load(char *s);
        class unit *findcarried(int transp);
        int getcarried(void) { return carried; }
        void setcarried(int c) { carried = c; }
        int getlfdnr(void) { return lfdnr; }
        void setlfdnr(int l) { lfdnr = l; }
        class unit *findlfdnr(int nr);
        int gethp(void) { return maxdamage()-damage; }
        int fortified(void) { return owner >= FORTIFY; }
        void fortify(void) { if (!fortified()) owner += FORTIFY; }
        void unfortify(void) { if (fortified()) owner -= FORTIFY; }
        char getowner(int compmove = 0) { if (!compmove) unfortify(); return fortified() ? owner - FORTIFY : owner; }
        int is_landunit(void) { return id() <= PANZER; }
    protected:
        int lfdnr;
        int hasmoved;
        int damage;
        int repaired;
        int carried;
    };

class landunit : public unit
{   public:
    landunit(list & thislist) : unit(thislist) {};
    landunit() : unit() {};
    virtual ~landunit() {}
    virtual int validsquare(coordinate c);
    virtual int canmove(void) { return 1; }
    virtual int movement_cost(coordinate dest);
    };

class infantrie : public landunit
{   public:
    infantrie(list & thislist) : landunit(thislist) {};
    infantrie() : landunit() {};
    virtual ~infantrie() {}
        virtual int id() { return INFANTRIE; }
	virtual char * name() { return T_ARMY; }
	virtual char prodtime() { return 6; }
        virtual int maxdamage(void) { return 1; }
        virtual int repairtime(void) { return 0; } // can't be repaired
    };

class panzer : public landunit
{   public:
        panzer(list & thislist) : landunit(thislist) {};
        panzer() : landunit() {};
        virtual ~panzer() {}
        virtual int id() { return PANZER; }
	virtual char * name() { return T_TANK; }
	virtual char prodtime() { return 8; }
        virtual int maxdamage(void) { return 2; }
        virtual int repairtime(void) { return 3; }
    };

class flieger : public unit
{   public:
        flieger(list & thislist) : unit(thislist) {};
        flieger() : unit() { fuel = 20; };
        virtual ~flieger() {}
        virtual char * restmovestring(char * s);
        virtual int id() { return FLIEGER; }
	virtual char * name() { return T_AIRCRAFT; }
	virtual char prodtime() { return 10; }
        virtual int validsquare(coordinate c);
        virtual int canmove(void) { return 5; }
        virtual int maxdamage(void) { return 1; }
        virtual int repairtime(void) { return 0; } // can't be repaired
        virtual void save(FILE *f);
        virtual void load(char *s);
        int should_return(void) { return fuel < 11; }
    private:
        int fuel;
    };

class ship : public unit
{   public:
    ship(list & thislist) : unit(thislist) {};
    ship() : unit() {};
    virtual ~ship() {}
    virtual int validsquare(coordinate c);
    };

class uboot : public ship
{   public:
    uboot(list & thislist) : ship(thislist) {};
    uboot() : ship() {};
    virtual ~uboot() {}
        virtual int id() { return UBOOT; }
	virtual char * name() { return T_SUBMARINE; }
	virtual char prodtime() { return 12; }
        virtual int canmove(void) { return 3; }
        virtual int maxdamage(void) { return 1; }
        virtual int repairtime(void) { return 0; } // can't be repaired
    };

class transporter : public ship
{   public:
    transporter(list & thislist) : ship(thislist) {};
    transporter() : ship() { nextunload = 0; };
    virtual ~transporter() {}
    void setnextunload(int u) { nextunload = u; }
    int getnextunload(void) { return nextunload; }
    virtual int validsquare(coordinate c);
        virtual int id() { return TRANSPORTER; }
	virtual char * name() { return T_FERRY; }
	virtual char prodtime() { return 10; }
        virtual int canmove(void) { return 2; }
        virtual int maxdamage(void) { return 2; }
        virtual int repairtime(void) { return 5; }
        virtual void save(FILE *f);
        virtual void load(char *s);
    private:
       int nextunload;
    };

class schnellboot : public ship
{   public:
    schnellboot(list & thislist) : ship(thislist) {};
    schnellboot() : ship() {};
    virtual ~schnellboot() {}
        virtual int id() { return SCHNELLBOOT; }
	virtual char * name() { return T_BOAT; }
	virtual char prodtime() { return 12; }
        virtual int canmove(void) { return 4; }
        virtual int maxdamage(void) { return 2; }
        virtual int repairtime(void) { return 4; }
    };

class fregatte : public ship
{   public:
    fregatte(list & thislist) : ship(thislist) {};
    fregatte() : ship() {};
    virtual ~fregatte() {}
        virtual int id() { return FREGATTE; }
	virtual char * name() { return T_FRIGATE; }
	virtual char prodtime() { return 15; }
        virtual int canmove(void) { return 3; }
        virtual int maxdamage(void) { return 4; }
        virtual int repairtime(void) { return 3; }
    };

class schlachtschiff : public ship
{   public:
    schlachtschiff(list & thislist) : ship(thislist) {};
    schlachtschiff() : ship () {};
    virtual ~schlachtschiff() {}
        virtual int id() { return SCHLACHTSCHIFF; }
	virtual char * name() { return T_BATTLESHIP; }
	virtual char prodtime() { return 22; }
        virtual int canmove(void) { return 2; }
        virtual int maxdamage(void) { return 7; }
        virtual int repairtime(void) { return 2; }
    };

class minenleger : public ship
{   public:
    minenleger(list & thislist) : ship(thislist) {};
    minenleger() : ship() {};
    virtual ~minenleger() {}
        virtual int id() { return MINENLEGER; }
	virtual char * name() { return T_MINEBOAT; }
	virtual char prodtime() { return 30; }
        virtual int canmove(void) { return 2; }
        virtual int maxdamage(void) { return 1; }
        virtual int repairtime(void) { return 0; } // can't be repaired
    };

class spieler
{   public:
	spieler();
	char name[40];
	int used, computer, ext;
        void save(FILE *f);
        void load(FILE *f);
    };

class game
{   public:
	game();
        list cities;
        list units;
	map *mp;
	matrix ownership;
	matrix unitmap, roadmap;
	spieler sp[7];
        void lookout(int sp); /* build seen-matrix and road directions*/
        void explore(coordinate c, int spieler);
	void initcities(void);
	int activeplayer;
        int unitcount;
        int load(char *file);
        int save(char *file);
        int bekannt(coordinate c);
        int bekannt(coordinate c, int sp);
        int beobachtet(coordinate c);
        int beobachtet(coordinate c, int sp);
        int erkundet(coordinate c) { return known[c] & (1 << activeplayer); }
        int meins(coordinate c) { return ownership[c] == activeplayer; }
        int feindlich(coordinate c) { return ownership[c] && !meins(c); }
        int feindgesichtet(coordinate c) { return feindlich(c) && beobachtet(c); }
        int wasser(coordinate c) { return (*mp)[c] < 3; }
        int coast(coordinate c);
        int stadt(coordinate c) { return (*mp)[c] == 5; }
        int strasse(coordinate c) { return (*mp)[c] == 6; }
        int freiland(coordinate c) { return (*mp)[c] == 3; }
        int fremde_stadt(coordinate c) { return stadt(c) && !meins(c); }
        int vermint(coordinate c);
        int minebekannt(coordinate c);
        void minelegen(coordinate c);
        void compmove(void);
    private:
        void move_planes(class plan *land);
        void move_land_units(class plan *land);
        void change_production(class plan *land);
        void move_transporters(class plan *land);
        void move_navy(class plan *land);
        matrix mine;
        matrix known;
        matrix seen;
    };

#endif
